﻿Imports System.Net
Imports System.Text
Imports System.IO
Imports System.Xml
Imports System.Security.Principal
Imports System.Security.Cryptography.X509Certificates
Imports System.Net.Security

Public Class frmMain
#Region "Constants"
    'Some constants defining the Globe environment
    Const SQL_SERVER = "(local)\SQL2012"
    Const SQL_DATABASE = "MacBean_GL"
    Const GLOBE_URL = "http://dijk177740-vpc2:8020/services/Exact.Entity.REST.EG/"

    'Namespaces that are used in the xml messages.
    Const NS_DATASERVICE = "http://schemas.microsoft.com/ado/2007/08/dataservices"
    Const NS_METADATA = "http://schemas.microsoft.com/ado/2007/08/dataservices/metadata"
#End Region

#Region "Help functions"
    'Certificate validation for https requests
    Private Function ValidateCertificate(sender As Object, certificate As X509Certificate, chain As X509Chain, sslPolicyErrors As SslPolicyErrors) As Boolean
        Return True
    End Function

    'Send the xml and return the response
    Private Function RequestAndResponse(ByVal xml As String, ByVal method As String, ByVal url As String) As String
        Dim httpRequest As HttpWebRequest
        Dim cc As CredentialCache
        Dim httpResponse As HttpWebResponse

        ServicePointManager.ServerCertificateValidationCallback = AddressOf ValidateCertificate

        'Set NTLM Authentication
        httpRequest = HttpWebRequest.Create(url)
        httpRequest.UseDefaultCredentials = False
        cc = New CredentialCache()
        cc.Add(New Uri(url), "NTLM", New NetworkCredential(tbDomain.Text & "\" & tbUsername.Text, tbPassword.Text))
        httpRequest.Credentials = cc

        'Add headers to login to the correct Globe company
        'Since Exact Globe Next 406 it is not needed anymore to pass credentials in the header
        'httpRequest.Headers.Add("Username", tbDomain.Text & "\" & tbUsername.Text)
        'httpRequest.Headers.Add("Password", Convert.ToBase64String(Encoding.ASCII.GetBytes(tbPassword.Text)))
        httpRequest.Headers.Add("ServerName", SQL_SERVER)
        httpRequest.Headers.Add("DatabaseName", SQL_DATABASE)

        'Set method and content type
        httpRequest.ContentType = "application/atom+xml"
        httpRequest.Method = method
        If cbJSON.Checked Then
            httpRequest.Accept = "application/atom+json,application/json; charset=utf-8"
        Else
            httpRequest.Accept = "application/atom+xml,application/xml; charset=utf-8"
        End If

        'Set the XML Message
        If xml <> "" Then
            Using s As Stream = httpRequest.GetRequestStream
                Using sw As New StreamWriter(s)
                    sw.Write(xml)
                End Using
            End Using
        End If

        'Get the response
        httpResponse = httpRequest.GetResponse()

        'Return the response message
        Using sr As New StreamReader(httpResponse.GetResponseStream, Encoding.GetEncoding("utf-8"))
            RequestAndResponse = sr.ReadToEnd()
        End Using

        ServicePointManager.ServerCertificateValidationCallback = Nothing
    End Function

    'Function to create the skeleton of the XML message
    Private Function CreateXMLDoc(ByVal entity As String, Optional ByVal id As String = "") As XmlDocument
        Dim doc As New XmlDocument()
        Dim root As XmlElement
        Dim element As XmlElement

        'Create the header
        doc.AppendChild(doc.CreateXmlDeclaration("1.0", "utf-8", "yes"))
        root = doc.CreateElement("entry")
        root.SetAttribute("xmlns:d", NS_DATASERVICE)
        root.SetAttribute("xmlns:m", NS_METADATA)
        root.SetAttribute("xmlns", "http://www.w3.org/2005/Atom")
        doc.AppendChild(root)

        'Create category, title, author, name, updated time and ID nodes
        element = doc.CreateElement("category")
        element.SetAttribute("scheme", "http://schemas.microsoft.com/ado/2007/08/dataservices/scheme")
        element.SetAttribute("term", entity)
        root.AppendChild(element)
        root.AppendChild(doc.CreateElement("title"))
        element = doc.CreateElement("author")
        element.AppendChild(doc.CreateElement("name"))
        root.AppendChild(element)
        element = doc.CreateElement("updated")
        element.InnerText = XmlConvert.ToString(Now, XmlDateTimeSerializationMode.Utc)
        root.AppendChild(element)
        If id <> "" Then
            element = doc.CreateElement("id")
            element.InnerText = id
            root.AppendChild(element)
        End If
        'Create content node

        element = doc.CreateElement("content")
        element.SetAttribute("type", "application/xml")
        root.AppendChild(element)

        Return doc
    End Function

    'Function to retrieve the value of a property
    Private Function GetPropertyValue(ByVal xml As String, ByVal p As String) As String
        Dim ns As XmlNamespaceManager
        Dim node As XmlNode
        Dim doc = New XmlDocument()

        'Load the xml into an XmlDocument
        doc.LoadXml(xml)

        'Add the three namespaces that are used
        ns = New XmlNamespaceManager(doc.NameTable)
        ns.AddNamespace("r", "http://www.w3.org/2005/Atom")
        ns.AddNamespace("d", NS_DATASERVICE)
        ns.AddNamespace("m", NS_METADATA)

        'Query the property node using an XPath expression
        node = doc.SelectSingleNode("//r:content/m:properties/" & p, ns)
        'node = doc.SelectSingleNode("/r:entry/r:content/m:properties/" & p, ns)

        'Return the value
        If Not node Is Nothing Then
            Return node.InnerText
        Else
            Return ""
        End If
    End Function

    'Function to retrieve the values of a certain property (eg. the line ID's)
    Private Function GetPropertyList(ByVal xml As String, ByVal p As String) As List(Of String)
        Dim ns As XmlNamespaceManager
        Dim node As XmlNode
        Dim nl As XmlNodeList
        Dim doc = New XmlDocument()
        Dim sl = New List(Of String)

        'Load the xml into an XmlDocument
        doc.LoadXml(xml)

        'Add the three namespaces that are used
        ns = New XmlNamespaceManager(doc.NameTable)
        ns.AddNamespace("r", "http://www.w3.org/2005/Atom")
        ns.AddNamespace("d", NS_DATASERVICE)
        ns.AddNamespace("m", NS_METADATA)

        'Query the property node using an XPath expression
        node = doc.SelectSingleNode("//r:content/m:properties/" & p, ns)

        'Process the node. Please note that the content is XML Encoded, so it must be parsed a second time.
        If Not node Is Nothing Then
            doc.LoadXml(node.InnerText)
            nl = doc.SelectNodes("//id")

            If Not nl Is Nothing Then
                For Each n As XmlNode In nl
                    sl.Add(n.InnerText)
                Next
            End If
        End If
        Return sl
    End Function

    'Function to retrieve the URL of the next record
    Private Function GetNextUrl(ByVal xml As String) As String
        Dim ns As XmlNamespaceManager
        Dim node As XmlNode
        Dim doc = New XmlDocument()

        'Load the xml into an XmlDocument
        doc.LoadXml(xml)

        'Add the three namespaces that are used
        ns = New XmlNamespaceManager(doc.NameTable)
        ns.AddNamespace("r", "http://www.w3.org/2005/Atom")
        ns.AddNamespace("d", NS_DATASERVICE)
        ns.AddNamespace("m", NS_METADATA)

        'Query the property node using an XPath expression
        node = doc.SelectSingleNode("/r:feed/r:link[@rel=""next""]/@href", ns)

        'Return the value
        If Not node Is Nothing Then
            Return node.InnerText
        Else
            Return ""
        End If
    End Function
#End Region

#Region "REST Service calls"
    'Sub to query a debtor (on debtor code)
    Private Sub RetrieveDebtor()
        Dim url As String = ""
        Dim response As String = ""
        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "Account/?$top=10&$filter=DebtorCode eq '               60089'&$orderby=DebtorCode"

            tbMessage.Text = url & vbCrLf

            'Call the Account REST service
            response = RequestAndResponse("", "GET", url)

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response
        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to retrieve a bankaccount
    Private Sub RetrieveBankAccount()
        Dim url As String = ""
        Dim response As String = ""
        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "BankAccount/?$top=5&$filter=BankAccountNumber eq '0058745874                        '"

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & url & vbCrLf

            'Call the Account REST service
            response = RequestAndResponse("", "GET", url)

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response
        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to retrieve a Financial entry
    Private Sub RetrieveFinancialEntry()
        Dim url As String = ""
        Dim response As String = ""
        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "FinancialHeader/?$top=1&$filter=(Journal eq '600') and (EntryNumber eq '13000034') and FinancialYear eq 2013"

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & url & vbCrLf

            'Call the Account REST service
            response = RequestAndResponse("", "GET", url)

            'Add the response to the text box
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & "Payment terms: "
            For Each s As String In GetPropertyList(response, "d:PaymentTermLines")
                'tbMessage.Text = tbMessage.Text & s & ", "
                RetrieveFinancialPaymentTerm(s)
            Next
            'tbMessage.Text = tbMessage.Text.Remove(tbMessage.Text.Length() - 2)

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & "Lines: "
            For Each s As String In GetPropertyList(response, "d:EntryLines")
                'tbMessage.Text = tbMessage.Text & s & ", "
                RetrieveFinancialLine(s)
            Next
            'tbMessage.Text = tbMessage.Text.Remove(tbMessage.Text.Length() - 2)

        Catch we As WebException
            If Not we.Response Is Nothing Then
                Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                    tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
                End Using
            End If
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to retrieve the line of a Financial entry
    Private Sub RetrieveFinancialLine(ByVal lineID As String)
        Dim url As String = ""
        Dim response As String = ""
        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "FinancialLine/?$top=1&$filter=ID eq " & lineID

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & url & vbCrLf

            'Call the Account REST service
            response = RequestAndResponse("", "GET", url)

            'Add the response to the text box
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response

        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to retrieve the line of a Financial entry
    Private Sub RetrieveFinancialPaymentTerm(ByVal lineID As String)
        Dim url As String = ""
        Dim response As String = ""
        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "FinancialPaymentTerm/?$top=1&$filter=ID eq " & lineID
            'url = GLOBE_URL & "FinancialPaymentTerm/?$top=1&$filter=DebtorNumber eq '60089'"
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & url & vbCrLf

            'Call the Account REST service
            response = RequestAndResponse("", "GET", url)

            'Add the response to the text box
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response

        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to retrieve the outstanding amount of a debtor
    Private Sub RetrieveOutstandingAmount()
        Dim url As String = ""
        Dim response As String = ""
        Dim outstandingAmount As Double = 0.0
        Dim s As String = ""
        Dim d As Double = 0.0

        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "FinancialHeader/?$top=1&$filter=(Journal eq ' 70') and (DebtorNumber eq ' 60089') and (FinancialYear eq 2013)"

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & url & vbCrLf

            'Call the FinancialHeader REST service
            response = RequestAndResponse("", "GET", url)

            'Retrieve and add the value of the outstanding amount
            s = GetPropertyValue(response, "d:OutstandingAmount")
            If Double.TryParse(s, d) Then
                outstandingAmount += d
            End If

            'Get the URL for the next record
            url = GetNextUrl(response)

            'Keep getting Financial Headers as long as there are available
            While url <> ""
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & url & vbCrLf

                'Call the Financial Header REST service
                response = RequestAndResponse("", "GET", url)

                'Retrieve and add the value of the outstanding amount
                s = GetPropertyValue(response, "d:OutstandingAmount")
                If Double.TryParse(s, d) Then
                    outstandingAmount += d
                End If

                'Get the URL for the next record
                url = GetNextUrl(response)
            End While

            'Add the outstanding amount to the text box
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & "Outstanding amount: " & outstandingAmount

        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to create a new bank account
    Private Sub CreateBankAccount()
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement
        Dim httpResponse As String

        Try
            tbMessage.Text = "Create Financial entry" & vbCrLf

            'This example does not support JSON for creating financial entries
            cbJSON.Checked = False

            'Create the basis XML document
            doc = CreateXMLDoc("Exact.Metadata.Entity.BankAccount")
            content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

            'Create the properties node
            properties = doc.CreateElement("m", "properties", NS_METADATA)

            element = doc.CreateElement("d", "AccountID", NS_DATASERVICE)
            element.InnerText = "{e11c4af3-7fa9-4a61-9152-4a8b7743f911}" 'GUID of the debtor
            properties.AppendChild(element)

            element = doc.CreateElement("d", "BankAccountType", NS_DATASERVICE)
            element.InnerText = "NL"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "BankAccountNumber", NS_DATASERVICE)
            element.InnerText = "0417164300EUR"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "IBAN", NS_DATASERVICE)
            element.InnerText = "NL91ABNA0417164300"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "BICCode", NS_DATASERVICE)
            element.InnerText = "ABNANL2A"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "AccountNumberBank", NS_DATASERVICE)
            element.InnerText = "0417164300"
            properties.AppendChild(element)

            'Add al properties to the content node of the XML message
            content.AppendChild(properties)

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & doc.OuterXml

            'Call the REST service with the XML message as payload
            httpResponse = RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "BankAccount")

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & httpResponse
        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to create a sales journal entry
    Private Sub CreateFinEntry()
        Dim transactionKey As String
        Try
            tbMessage.Text = "Create Financial entry" & vbCrLf

            'This example does not support JSON for creating financial entries
            cbJSON.Checked = False

            'Create the line and retrieve the transaction key
            transactionKey = CreateFinLine("", "70", "8000", "Line 1", 100, "2")

            'Create the payment term
            CreateFinPaymentTerm(transactionKey, 119, "60089", "02", "I", "Payment term 1", "", "")

            'Create the header (please be aware that my demo database still uses 19% VAT)
            CreateFinHeader(transactionKey, "70", "60089", "Financial entry 1", 119, "02")

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & transactionKey
        Catch we As WebException
            If Not we.Response Is Nothing Then
                Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                    tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
                End Using
            Else
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Message
            End If
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Function to create a sales journal entry line, it retrieves the transaction key
    Private Function CreateFinLine(ByVal transactionKey As String, ByVal journal As String, ByVal generalLedger As String, ByVal description As String, ByVal amount As Double, ByVal vat As String) As String
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement
        Dim httpResponse As String

        'Create the basis XML document
        doc = CreateXMLDoc("Exact.Metadata.Entity.FinancialLine")
        content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

        'Create the properties node
        properties = doc.CreateElement("m", "properties", NS_METADATA)

        'Add all properies
        element = doc.CreateElement("d", "ID", NS_DATASERVICE)
        element.InnerText = "0" 'Is mandatory but will not be used in the database, so alway pass 0
        properties.AppendChild(element)

        'Only add the transaction key if it is provided.
        If transactionKey <> "" Then
            element = doc.CreateElement("d", "TransactionKey", NS_DATASERVICE)
            element.InnerText = transactionKey
            properties.AppendChild(element)
        End If

        element = doc.CreateElement("d", "Journal", NS_DATASERVICE)
        element.InnerText = journal
        properties.AppendChild(element)

        element = doc.CreateElement("d", "GLAccount", NS_DATASERVICE)
        element.InnerText = generalLedger
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Description", NS_DATASERVICE)
        element.InnerText = description
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Amount", NS_DATASERVICE)
        element.InnerText = amount.ToString("F", Globalization.CultureInfo.CreateSpecificCulture("en-US"))

        properties.AppendChild(element)

        element = doc.CreateElement("d", "VATCode", NS_DATASERVICE)
        element.InnerText = vat
        properties.AppendChild(element)

        'Add al properties to the content node of the XML message
        content.AppendChild(properties)

        'Call the REST service with the XML message as payload
        httpResponse = RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "FinancialLine")

        'Return the transaction key
        Return GetPropertyValue(httpResponse, "d:TransactionKey")
    End Function

    'Sub to create a payment term
    Private Sub CreateFinPaymentTerm(ByVal transactionKey As String, ByVal amount As Double, ByVal debtor As String, _
                                     ByVal method As String, ByVal type As String, ByVal description As String, ByVal ownBankNumber As String, ByVal debtorBankNumber As String)
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement

        'Create the basis XML document
        doc = CreateXMLDoc("Exact.Metadata.Entity.FinancialPaymentTerm")
        content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

        'Create the properties node
        properties = doc.CreateElement("m", "properties", NS_METADATA)

        'Add all properies
        If transactionKey <> "" Then
            element = doc.CreateElement("d", "TransactionKey", NS_DATASERVICE)
            element.InnerText = transactionKey
            properties.AppendChild(element)
        End If

        element = doc.CreateElement("d", "ID", NS_DATASERVICE)
        element.InnerText = "0" 'Is mandatory but will not be used in the database, so alway pass 0
        properties.AppendChild(element)

        element = doc.CreateElement("d", "AmountDC", NS_DATASERVICE)
        element.InnerText = amount.ToString("F", Globalization.CultureInfo.CreateSpecificCulture("en-US"))
        properties.AppendChild(element)

        element = doc.CreateElement("d", "DebtorNumber", NS_DATASERVICE)
        element.InnerText = debtor
        properties.AppendChild(element)

        element = doc.CreateElement("d", "AmountInTransactionCurrency", NS_DATASERVICE)
        element.InnerText = amount.ToString("F", Globalization.CultureInfo.CreateSpecificCulture("en-US"))
        properties.AppendChild(element)

        element = doc.CreateElement("d", "TransactionCurrencyCode", NS_DATASERVICE)
        element.InnerText = "EUR"
        properties.AppendChild(element)

        element = doc.CreateElement("d", "PaymentCondition", NS_DATASERVICE)
        element.InnerText = method
        properties.AppendChild(element)

        element = doc.CreateElement("d", "PaymentType", NS_DATASERVICE)
        element.InnerText = type
        properties.AppendChild(element)

        element = doc.CreateElement("d", "TransactionType", NS_DATASERVICE)
        element.InnerText = "K" 'Sales invoice
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Status", NS_DATASERVICE)
        element.InnerText = "C" 'Entered
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Type", NS_DATASERVICE)
        element.InnerText = "W" 'Installment
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Description", NS_DATASERVICE)
        element.InnerText = description
        properties.AppendChild(element)

        element = doc.CreateElement("d", "OwnBankAccount", NS_DATASERVICE)
        element.InnerText = "0118292161EUR" 'this must be the value from the database field: BankAccounts.BankAccount
        properties.AppendChild(element)

        element = doc.CreateElement("d", "OffsetBankAccount", NS_DATASERVICE)
        element.InnerText = "0118292161" 'this must be the value from the database field: BnkAcc.Banknr
        properties.AppendChild(element)

        content.AppendChild(properties)

        'Call the REST service with the XML message as payload
        RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "FinancialPaymentTerm")
    End Sub

    'Sub to create a financial header
    Private Sub CreateFinHeader(ByVal transactionKey As String, ByVal journal As String, ByVal debtor As String, ByVal description As String, ByVal amount As Double, ByVal paycondition As String)
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement

        'Create the basis XML document
        doc = CreateXMLDoc("Exact.Metadata.Entity.FinancialHeader")
        content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

        'Create the properties node
        properties = doc.CreateElement("m", "properties", NS_METADATA)

        'Add all properies
        If transactionKey <> "" Then
            element = doc.CreateElement("d", "TransactionKey", NS_DATASERVICE)
            element.InnerText = transactionKey
            properties.AppendChild(element)
        End If

        element = doc.CreateElement("d", "Journal", NS_DATASERVICE)
        element.InnerText = journal
        properties.AppendChild(element)

        element = doc.CreateElement("d", "FinancialYear", NS_DATASERVICE)
        element.InnerText = Now.Year
        properties.AppendChild(element)

        element = doc.CreateElement("d", "FinancialPeriod", NS_DATASERVICE)
        element.InnerText = Now.Month.ToString().PadLeft(3, "0")
        properties.AppendChild(element)

        element = doc.CreateElement("d", "DebtorNumber", NS_DATASERVICE)
        element.InnerText = debtor
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Description", NS_DATASERVICE)
        element.InnerText = description
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Amount", NS_DATASERVICE)
        element.InnerText = amount.ToString("F", Globalization.CultureInfo.CreateSpecificCulture("en-US"))
        properties.AppendChild(element)

        element = doc.CreateElement("d", "PaymentCondition", NS_DATASERVICE)
        element.InnerText = paycondition
        properties.AppendChild(element)


        content.AppendChild(properties)

        'Call the REST service with the XML message as payload
        RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "FinancialHeader")
    End Sub

    'Sub to create a sales order entry
    Private Sub CreateSalesOrder()
        Dim transactionKey As String
        Try
            tbMessage.Text = "Create Sales order" & vbCrLf

            'This example does not support JSON for creating financial entries
            cbJSON.Checked = False

            'Create the line and retrieve the transaction key
            transactionKey = CreateSOLine("")
            'transactionKey = CreateFinLine("", "700", "800000", "Line 1", 100.5, "0") '"VHI")

            'Create the header (please be aware that my demo database still uses 19% VAT)
            CreateSOHeader(transactionKey)
            'CreateFinHeader(transactionKey, "700", "10227", "Financial entry 1", 100.5, "30")

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & transactionKey
        Catch we As WebException
            If Not we.Response Is Nothing Then
                Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                    tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
                End Using
            Else
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Message
            End If
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Function to create a sales order line, it retrieves the transaction key
    Private Function CreateSOLine(ByVal transactionKey As String) As String
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement
        Dim httpResponse As String

        'Create the basis XML document
        doc = CreateXMLDoc("Exact.Metadata.Entity.SalesOrderLine")
        content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

        'Create the properties node
        properties = doc.CreateElement("m", "properties", NS_METADATA)

        'Add all properies
        element = doc.CreateElement("d", "ID", NS_DATASERVICE)
        element.InnerText = "0" 'Is mandatory but will not be used in the database, so alway pass 0
        properties.AppendChild(element)

        'Only add the transaction key if it is provided.
        If transactionKey <> "" Then
            element = doc.CreateElement("d", "TransactionKey", NS_DATASERVICE)
            element.InnerText = transactionKey
            properties.AppendChild(element)
        End If

        element = doc.CreateElement("d", "Itemcode", NS_DATASERVICE)
        element.InnerText = "BEK0001"
        properties.AppendChild(element)

        element = doc.CreateElement("d", "Quantity", NS_DATASERVICE)
        element.InnerText = 1
        properties.AppendChild(element)

        'Add al properties to the content node of the XML message
        content.AppendChild(properties)

        'Call the REST service with the XML message as payload
        httpResponse = RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "SalesOrderLine")

        'Return the transaction key
        Return GetPropertyValue(httpResponse, "d:TransactionKey")
    End Function

    'Sub to create a sales order header
    Private Sub CreateSOHeader(ByVal transactionKey As String)
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement

        'Create the basis XML document
        doc = CreateXMLDoc("Exact.Metadata.Entity.SalesOrderHeader")
        content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

        'Create the properties node
        properties = doc.CreateElement("m", "properties", NS_METADATA)

        'Add all properies
        If transactionKey <> "" Then
            element = doc.CreateElement("d", "TransactionKey", NS_DATASERVICE)
            element.InnerText = transactionKey
            properties.AppendChild(element)
        End If

        element = doc.CreateElement("d", "YourReference", NS_DATASERVICE)
        element.InnerText = "Test webservice"
        properties.AppendChild(element)

        element = doc.CreateElement("d", "OrderDebtor", NS_DATASERVICE)
        element.InnerText = "60089"
        properties.AppendChild(element)

        content.AppendChild(properties)

        'Call the REST service with the XML message as payload
        RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "SalesOrderHeader")
    End Sub

    'Sub to create an hour entry
    Private Sub CreateHourEntry()
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement
        Dim httpResponse As String

        Try
            tbMessage.Text = "Create Hour entry" & vbCrLf

            'This example does not support JSON for creating financial entries
            cbJSON.Checked = False

            'Create the basis XML document
            doc = CreateXMLDoc("Exact.Metadata.Entity.HourEntryLine")
            content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

            'Create the properties node
            properties = doc.CreateElement("m", "properties", NS_METADATA)

            element = doc.CreateElement("d", "ProjectCode", NS_DATASERVICE)
            element.InnerText = "20023" '               "
            properties.AppendChild(element)

            element = doc.CreateElement("d", "ItemCode", NS_DATASERVICE)
            element.InnerText = "ALG"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "Activity", NS_DATASERVICE)
            element.InnerText = "General hours"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "ResourceID", NS_DATASERVICE)
            element.InnerText = "1000"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "LineEntryDate", NS_DATASERVICE)
            element.InnerText = XmlConvert.ToString(Now.Date, XmlDateTimeSerializationMode.Utc)
            properties.AppendChild(element)

            element = doc.CreateElement("d", "Option", NS_DATASERVICE)
            element.InnerText = "PSA"
            properties.AppendChild(element)

            element = doc.CreateElement("d", "StartTime", NS_DATASERVICE)
            element.InnerText = XmlConvert.ToString(Now.Date, XmlDateTimeSerializationMode.Utc)
            properties.AppendChild(element)

            element = doc.CreateElement("d", "EndTime", NS_DATASERVICE)
            element.InnerText = XmlConvert.ToString(Now.Date, XmlDateTimeSerializationMode.Utc)
            properties.AppendChild(element)

            element = doc.CreateElement("d", "ActualQuantity", NS_DATASERVICE)
            element.InnerText = "1.0"
            properties.AppendChild(element)

            'Add al properties to the content node of the XML message
            content.AppendChild(properties)

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & doc.OuterXml

            'Call the REST service with the XML message as payload
            httpResponse = RequestAndResponse(doc.OuterXml, "POST", GLOBE_URL & "HourEntryLine")

            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & httpResponse
        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try
    End Sub

    'Sub to update a debtor
    Private Sub UpdateDebtor(ByVal debtorCode As String)
        Dim url As String = ""
        Dim response As String = ""
        Dim doc As XmlDocument
        Dim content As XmlElement
        Dim element As XmlElement
        Dim properties As XmlElement
        Dim id As String

        Try
            'Add the open data query options to the URL
            url = GLOBE_URL & "Account/?$top=10&$filter=DebtorCode eq '" & debtorCode & "'&$orderby=DebtorCode"

            tbMessage.Text = url & vbCrLf

            'Call the Account REST service
            response = RequestAndResponse("", "GET", url)
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response

            'get the ID
            id = GetPropertyValue(response, "d:ID")

            'Create a new xml document with the properties to change
            doc = CreateXMLDoc("Exact.Metadata.Entity.Account")
            content = CType(doc.SelectSingleNode("/entry/content"), XmlElement)

            'Create the properties node
            properties = doc.CreateElement("m", "properties", NS_METADATA)

            element = doc.CreateElement("d", "TextFreeField1", NS_DATASERVICE)
            element.InnerText = "updated: " & Now().ToUniversalTime()
            properties.AppendChild(element)

            content.AppendChild(properties)

            'merge the changes
            url = GLOBE_URL & "Account(guid'" & id & "')"
            response = RequestAndResponse(doc.OuterXml, "MERGE", url)
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & response
        Catch we As WebException
            Using reader As New StreamReader(we.Response.GetResponseStream, Encoding.UTF8)
                tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & we.Response.ToString() & vbCrLf & vbCrLf & reader.ReadToEnd()
            End Using
        Catch ex As Exception
            tbMessage.Text = tbMessage.Text & vbCrLf & vbCrLf & ex.Message
        End Try

    End Sub
#End Region

#Region "Events"
    Private Sub btnRetrieve_Click(sender As Object, e As EventArgs) Handles btnRetrieve.Click
        'RetrieveDebtor()
        'RetrieveBankAccount()
        'RetrieveOutstandingAmount()
        RetrieveFinancialEntry()
    End Sub

    Private Sub btnCreate_Click(sender As Object, e As EventArgs) Handles btnCreate.Click
        'CreateBankAccount()
        CreateFinEntry()
        'CreateHourEntry()
        'CreateSalesOrder()
    End Sub

    Private Sub btnUpdate_Click(sender As Object, e As EventArgs) Handles btnUpdate.Click
        UpdateDebtor("               60089")
    End Sub

    Private Sub btnClear_Click(sender As Object, e As EventArgs) Handles btnClear.Click
        tbMessage.Text = ""
    End Sub

    Private Sub frmMain_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        Dim identity As WindowsIdentity = WindowsIdentity.GetCurrent()
        Dim sl As String() = identity.Name.Split("\")
        If sl.Length > 1 Then
            tbDomain.Text = sl(0)
            tbUsername.Text = sl(1)
        Else
            tbUsername.Text = identity.Name
        End If
    End Sub
#End Region

End Class
